@@ -12,7 +12,7 @@
 # Copyright (c) 2018 Tim 'mithro' Ansell <me@mith.ro>
 # Copyright (c) 2018 Daniel Kucera <daniel.kucera@gmail.com>
 # Copyright (c) 2018 Mikołaj Sowiński <mikolaj.sowinski@gmail.com>
-# Copyright (c) 2020 Antmicro <www.antmicro.com>
+# Copyright (c) 2020-2021 Antmicro <www.antmicro.com>
 # SPDX-License-Identifier: BSD-2-Clause
 
 from math import ceil
@@ -294,6 +294,12 @@ def parse_spd_hexdump(filename):
 
 # SDRAMModule --------------------------------------------------------------------------------------
 
+Frac = namedtuple("Frac", ["num", "denom"])
+
+class Timing(namedtuple("Timing", ["ck", "ns"])):
+    def __add__(self, other):
+        return Timing(self.ck + other.ck, self.ns + other.ns)
+
 class SDRAMModule:
     """SDRAM module geometry and timings.
 
@@ -320,68 +326,82 @@ class SDRAMModule:
         if (fine_refresh_mode is None) and (self.memtype == "DDR4"):
             fine_refresh_mode = "1x"
         self.timing_settings = TimingSettings(
-            tRP   = self.ns_to_cycles(self.get("tRP")),
-            tRCD  = self.ns_to_cycles(self.get("tRCD")),
-            tWR   = self.ns_to_cycles(self.get("tWR")),
-            tREFI = self.ns_to_cycles(self.get("tREFI", fine_refresh_mode), False),
-            tRFC  = self.ck_ns_to_cycles(*self.get("tRFC", fine_refresh_mode)),
-            tWTR  = self.ck_ns_to_cycles(*self.get("tWTR")),
-            tFAW  = None if self.get("tFAW") is None else self.ck_ns_to_cycles(*self.get("tFAW")),
-            tCCD  = None if self.get("tCCD") is None else self.ck_ns_to_cycles(*self.get("tCCD")),
-            tRRD  = None if self.get("tRRD") is None else self.ck_ns_to_cycles(*self.get("tRRD")),
-            tRC   = None  if self.get("tRAS") is None else self.ns_to_cycles(self.get("tRP") + self.get("tRAS")),
-            tRAS  = None if self.get("tRAS") is None else self.ns_to_cycles(self.get("tRAS")),
-            tZQCS = None if self.get("tZQCS") is None else self.ck_ns_to_cycles(*self.get("tZQCS"))
+            tRP   = self.ck_ns_to_cycles(self.get("tRP")),
+            tRCD  = self.ck_ns_to_cycles(self.get("tRCD")),
+            tWR   = self.ck_ns_to_cycles(self.get("tWR")),
+            tREFI = self.ck_ns_to_cycles(self.get("tREFI", fine_refresh_mode), margin=False),
+            tRFC  = self.ck_ns_to_cycles(self.get("tRFC", fine_refresh_mode)),
+            tWTR  = self.ck_ns_to_cycles(self.get("tWTR")),
+            tFAW  = None if self.get("tFAW") is None else self.ck_ns_to_cycles(self.get("tFAW")),
+            tCCD  = None if self.get("tCCD") is None else self.ck_ns_to_cycles(self.get("tCCD")),
+            tRRD  = None if self.get("tRRD") is None else self.ck_ns_to_cycles(self.get("tRRD")),
+            tRC   = None if self.get("tRAS") is None else self.ck_ns_to_cycles(self.get("tRP") + self.get("tRAS")),
+            tRAS  = None if self.get("tRAS") is None else self.ck_ns_to_cycles(self.get("tRAS")),
+            tZQCS = None if self.get("tZQCS") is None else self.ck_ns_to_cycles(self.get("tZQCS"))
         )
         self.timing_settings.fine_refresh_mode = fine_refresh_mode
 
     def get(self, name, key=None):
-        r = None
-        if name in _speedgrade_timings:
-            if hasattr(self, "speedgrade_timings"):
-                speedgrade = "default" if self.speedgrade is None else self.speedgrade
-                r = getattr(self.speedgrade_timings[speedgrade], name)
-            else:
-                name = name + "_" + self.speedgrade if self.speedgrade is not None else name
-                try:
-                    r = getattr(self, name)
-                except:
-                    pass
+        assert name in _speedgrade_timings + _technology_timings, "Unknown name: {}".format(name)
+        timing = self.get_timing(name)
+        if timing is None:
+            return None
+        if (timing is not None) and (key is not None):
+            timing = timing[key]
+        if isinstance(timing, tuple):
+            ck, ns = timing
         else:
-            if hasattr(self, "technology_timings"):
-                r = getattr(self.technology_timings, name)
-            else:
-                try:
-                    r = getattr(self, name)
-                except:
-                    pass
-        if (r is not None) and (key is not None):
-            r = r[key]
-        return r
+            ck, ns = 0, timing
+        ck = ck or 0
+        ns = ns or 0
+        return Timing(ck, ns)
+
+    def get_timing(self, name):
+        if name in _speedgrade_timings:
+            return self.get_speedgrade_timing(name)
+        return self.get_technology_timing(name)
+
+    def get_speedgrade_timing(self, name):
+        if hasattr(self, "speedgrade_timings"):
+            speedgrade = "default" if self.speedgrade is None else self.speedgrade
+            return getattr(self.speedgrade_timings[speedgrade], name)
+        name = name + "_" + self.speedgrade if self.speedgrade is not None else name
+        try:
+            return getattr(self, name)
+        except:
+            pass
+
+    def get_technology_timing(self, name):
+        if hasattr(self, "technology_timings"):
+            return getattr(self.technology_timings, name)
+        else:
+            try:
+                return getattr(self, name)
+            except:
+                pass
 
     def ns_to_cycles(self, t, margin=True):
         clk_period_ns = 1e9/self.clk_freq
-        if margin:
-            margins = {
-                "1:1" : 0,
-                "1:2" : clk_period_ns/2,
-                "1:4" : 3*clk_period_ns/4
-            }
-            t += margins[self.rate]
+        t += self.margin if margin else 0
         return ceil(t/clk_period_ns)
 
     def ck_to_cycles(self, c):
-        d = {
-            "1:1" : 1,
-            "1:2" : 2,
-            "1:4" : 4
-        }
-        return ceil(c/d[self.rate])
+        return ceil(c/self.rate_frac.denom)
 
-    def ck_ns_to_cycles(self, c, t):
-        c = 0 if c is None else c
-        t = 0 if t is None else t
-        return max(self.ck_to_cycles(c), self.ns_to_cycles(t))
+    def ck_ns_to_cycles(self, timing, **kwargs):
+        return max(self.ck_to_cycles(timing.ck), self.ns_to_cycles(timing.ns, **kwargs))
+
+    @property
+    def rate_frac(self):
+        num, denom = map(int, self.rate.split(":"))
+        assert num == 1, "Rate: numerator must be 1: {}".format(self.rate)
+        return Frac(num, denom)
+
+    @property
+    def margin(self):
+        clk_period_ns = 1e9 / self.clk_freq
+        frac = self.rate_frac
+        return clk_period_ns * (1 - frac.num/frac.denom)
 
     @classmethod
     def from_spd_data(cls, spd_data, clk_freq, fine_refresh_mode=None):
@@ -461,6 +481,15 @@ class MT48LC16M16(SDRModule):
     technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=(None, 15))
     speedgrade_timings = {"default": _SpeedgradeTimings(tRP=20, tRCD=20, tWR=15, tRFC=(None, 66), tFAW=None, tRAS=44)}
 
+class MT48LC32M8(SDRModule):
+    # geometry
+    nbanks = 4
+    nrows  = 8192
+    ncols  = 1024
+    # timings
+    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=(None, 15))
+    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=20, tRCD=20, tWR=15, tRFC=(None, 66), tFAW=None, tRAS=44)}
+
 class AS4C16M16(SDRModule):
     # geometry
     nbanks = 4
@@ -604,6 +633,20 @@ class MT41K64M16(DDR3Module):
     }
     speedgrade_timings["default"] = speedgrade_timings["1600"]
 
+class MT41J256M8(DDR3Module):
+    # geometry
+    nbanks = 8
+    nrows  = 32768
+    ncols  = 1024
+    # timings
+    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 6), tZQCS=(64, 80))
+    speedgrade_timings = {
+        "800":  _SpeedgradeTimings(tRP=13.1,  tRCD=13.1,  tWR=15,  tRFC=(64, None),  tFAW=(None, 50), tRAS=37.5),
+        "1066": _SpeedgradeTimings(tRP=13.1,  tRCD=13.1,  tWR=15,  tRFC=(86, None),  tFAW=(None, 50), tRAS=37.5),
+        "1333": _SpeedgradeTimings(tRP=13.5,  tRCD=13.5,  tWR=15,  tRFC=(107, None), tFAW=(None, 45), tRAS=36),
+    }
+    speedgrade_timings["default"] = speedgrade_timings["1333"]
+
 class MT41J128M16(DDR3Module):
     # geometry
     nbanks = 8
@@ -706,6 +749,18 @@ class IS43TR16128B(DDR3Module):
     }
     speedgrade_timings["default"] = speedgrade_timings["1600"]
 
+class IS43TR16256A(DDR3Module):
+    # geometry
+    nbanks = 8
+    nrows  = 32768
+    ncols  = 1024
+    # timings
+    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 6), tZQCS=(64, 80))
+    speedgrade_timings = {
+        "1600": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=15, tRFC=(None, 260), tFAW=(None, 30), tRAS=35),
+    }
+    speedgrade_timings["default"] = speedgrade_timings["1600"]
+
 
 # DDR3 (SO-DIMM) -----------------------------------------------------------------------------------
 
@@ -923,3 +978,24 @@ class MTA18ASF2G72PZ(DDR4RegisteredModule):
         "2400": _SpeedgradeTimings(tRP=13.32, tRCD=13.32, tWR=15, tRFC=trfc, tFAW=(20, 25), tRAS=32),
     }
     speedgrade_timings["default"] = speedgrade_timings["2400"]
+
+# LPDDR4 -------------------------------------------------------------------------------------------
+
+class MT53E256M16D1(SDRAMModule):
+    memtype = "LPDDR4"
+
+    nbanks = 8
+    nrows = 32768
+    ncols = 1024
+
+    # TODO: find a way to select if we need masked writes
+    tccd = {"write": (8, None), "masked-write": (32, None)}
+
+    # TODO: tZQCS - performing ZQC during runtime will require modifying Refresher, as ZQC has to be done in 2 phases
+    # 1. ZQCAL START is issued 2. ZQCAL LATCH updates the values, the time START->LATCH tZQCAL=1us, so we cannot block
+    # the controller during this time, after ZQCAL LATCH we have to wait tZQLAT=max(8ck, 30ns)
+    technology_timings = _TechnologyTimings(tREFI=32e6/8192, tWTR=(8, 10), tCCD=tccd["masked-write"], tRRD=(4, 10), tZQCS=None)
+    speedgrade_timings = {
+        "1866": _SpeedgradeTimings(tRP=(3, 21), tRCD=(4, 18), tWR=(4, 18), tRFC=180, tFAW=40, tRAS=(3, 42)),  # TODO: tRAS_max
+    }
+    speedgrade_timings["default"] = speedgrade_timings["1866"]
