@@ -112,7 +112,13 @@ class ECP5DDRPHYInit(Module):
 # Lattice ECP5 DDR PHY -----------------------------------------------------------------------------
 
 class ECP5DDRPHY(Module, AutoCSR):
-    def __init__(self, pads, sys_clk_freq=100e6, dm_remapping={}):
+    def __init__(self, pads,
+        sys_clk_freq = 100e6,
+        cl           = None,
+        cwl          = None,
+        cmd_delay    = 0,
+        clk_polarity = 0):
+        assert isinstance(cmd_delay, int) and cmd_delay < 128
         pads        = PHYPadsCombiner(pads)
         memtype     = "DDR3"
         tck         = 2/(2*2*sys_clk_freq)
@@ -127,7 +133,8 @@ class ECP5DDRPHY(Module, AutoCSR):
         self.submodules.init = ECP5DDRPHYInit()
 
         # Parameters -------------------------------------------------------------------------------
-        cl, cwl         = get_cl_cw(memtype, tck)
+        cl  = get_default_cl( memtype, tck) if cl  is None else cl
+        cwl = get_default_cwl(memtype, tck) if cwl is None else cwl
         cl_sys_latency  = get_sys_latency(nphases, cl)
         cwl_sys_latency = get_sys_latency(nphases, cwl)
 
@@ -164,7 +171,7 @@ class ECP5DDRPHY(Module, AutoCSR):
         )
 
         # DFI Interface ----------------------------------------------------------------------------
-        self.dfi = dfi = Interface(addressbits, bankbits, nranks, 4*databits, 4)
+        self.dfi = dfi = Interface(addressbits, bankbits, nranks, 4*databits, nphases)
 
         # # #
 
@@ -175,37 +182,54 @@ class ECP5DDRPHY(Module, AutoCSR):
             pads.sel_group(pads_group)
 
             # Clock --------------------------------------------------------------------------------
+            clk_pattern = {0: 0b1010, 1: 0b0101}[clk_polarity]
             for i in range(len(pads.clk_p)):
-                sd_clk_se = Signal()
+                pad_oddrx2f = Signal()
                 self.specials += Instance("ODDRX2F",
                     i_RST  = ResetSignal("sys"),
                     i_SCLK = ClockSignal("sys"),
                     i_ECLK = ClockSignal("sys2x"),
-                    **{f"i_D{n}": (0b1010 >> n) & 0b1 for n in range(4)},
-                    o_Q    = pads.clk_p[i]
+                    **{f"i_D{n}": (clk_pattern >> n) & 0b1 for n in range(4)},
+                    o_Q    = pad_oddrx2f
+                )
+                self.specials += Instance("DELAYG",
+                    p_DEL_VALUE = cmd_delay,
+                    i_A         = pad_oddrx2f,
+                    o_Z         = pads.clk_p[i]
                 )
 
             # Commands -----------------------------------------------------------------------------
             commands = {
-                "a"    : "address",
-                "ba"   : "bank"   ,
-                "ras_n": "ras_n"  ,
-                "cas_n": "cas_n"  ,
-                "we_n" : "we_n"   ,
-                "cke"  : "cke"    ,
-                "odt"  : "odt"    ,
+                # Pad name: (DFI name,   Pad type (required or optional))
+                "reset_n" : ("reset_n", "optional"),
+                "cs_n"    : ("cs_n",    "optional"),
+                "a"       : ("address", "required"),
+                "ba"      : ("bank"   , "required"),
+                "ras_n"   : ("ras_n"  , "required"),
+                "cas_n"   : ("cas_n"  , "required"),
+                "we_n"    : ("we_n"   , "required"),
+                "cke"     : ("cke"    , "optional"),
+                "odt"     : ("odt"    , "optional"),
             }
-            if hasattr(pads, "reset_n"): commands.update({"reset_n" : "reset_n"})
-            if hasattr(pads, "cs_n")   : commands.update({"cs_n"    : "cs_n"})
-            for pad_name, dfi_name in commands.items():
-                pad = getattr(pads, pad_name)
+            for pad_name, (dfi_name, pad_type) in commands.items():
+                pad = getattr(pads, pad_name, None)
+                if (pad is None):
+                    if (pad_type == "required"):
+                        raise ValueError(f"DRAM pad {pad_name} required but not found in pads.")
+                    continue
                 for i in range(len(pad)):
+                    pad_oddrx2f = Signal()
                     self.specials += Instance("ODDRX2F",
                         i_RST  = ResetSignal("sys"),
                         i_SCLK = ClockSignal("sys"),
                         i_ECLK = ClockSignal("sys2x"),
                         **{f"i_D{n}": getattr(dfi.phases[n//2], dfi_name)[i] for n in range(4)},
-                        o_Q    = pad[i]
+                        o_Q    = pad_oddrx2f
+                    )
+                    self.specials += Instance("DELAYG",
+                        p_DEL_VALUE = cmd_delay,
+                        i_A         = pad_oddrx2f,
+                        o_Z         = pad[i]
                     )
 
         # DQS/DM/DQ --------------------------------------------------------------------------------
@@ -300,7 +324,7 @@ class ECP5DDRPHY(Module, AutoCSR):
             dm_o_data_d     = Signal(8)
             dm_o_data_muxed = Signal(4)
             for n in range(8):
-                self.comb += dm_o_data[n].eq(dfi.phases[n//4].wrdata_mask[n%4*databits//8+dm_remapping.get(i, i)])
+                self.comb += dm_o_data[n].eq(dfi.phases[n//4].wrdata_mask[n%4*databits//8 + i])
             self.sync += dm_o_data_d.eq(dm_o_data)
             dm_bl8_cases = {}
             dm_bl8_cases[0] = dm_o_data_muxed.eq(dm_o_data[:4])
@@ -348,13 +372,10 @@ class ECP5DDRPHY(Module, AutoCSR):
                     cycles = 1)
                 self.submodules += dq_i_bitslip
                 self.specials += [
-                    Instance("DELAYF",
-                        p_DEL_MODE  = "DQS_ALIGNED_X2",
-                        i_LOADN     = 0,
-                        i_MOVE      = 0,
-                        i_DIRECTION = 0,
-                        i_A         = dq_i,
-                        o_Z         = dq_i_delayed
+                    Instance("DELAYG",
+                        p_DEL_MODE = "DQS_ALIGNED_X2",
+                        i_A        = dq_i,
+                        o_Z        = dq_i_delayed
                     ),
                     Instance("IDDRX2DQA",
                         i_RST     = ResetSignal("sys"),
