@@ -7,7 +7,7 @@
 
 # 1:4, 1:2 frequency-ratio DDR2/DDR3 PHY for Xilinx's Series7
 # DDR2: 400, 533, 667, 800 and 1066 MT/s
-# DDR3: 800, 1066, 1333 and 1600 MT/s
+# DDR3: 800, 1066, 1333 , 1600 and 1866 MT/s
 
 from functools import reduce
 from operator import or_
@@ -29,6 +29,8 @@ class S7DDRPHY(Module, AutoCSR):
         nphases          = 4,
         sys_clk_freq     = 100e6,
         iodelay_clk_freq = 200e6,
+        cl               = None,
+        cwl              = None,
         cmd_latency      = 0,
         cmd_delay        = None):
         assert not (memtype == "DDR3" and nphases == 2)
@@ -50,7 +52,8 @@ class S7DDRPHY(Module, AutoCSR):
         }
         half_sys8x_taps = math.floor(tck/(4*iodelay_tap_average[iodelay_clk_freq]))
 
-        cl, cwl         = get_cl_cw(memtype, tck)
+        cl              = get_default_cl( memtype, tck) if cl  is None else cl
+        cwl             = get_default_cwl(memtype, tck) if cwl is None else cwl
         cl_sys_latency  = get_sys_latency(nphases, cl)
         cwl_sys_latency = get_sys_latency(nphases, cwl)
         rdphase         = get_sys_phase(nphases, cl_sys_latency,   cl + cmd_latency)
@@ -126,9 +129,9 @@ class S7DDRPHY(Module, AutoCSR):
                     p_TRISTATE_WIDTH = 1,
                     p_DATA_RATE_OQ   = "DDR",
                     p_DATA_RATE_TQ   = "BUF",
-                    i_RST    = ResetSignal() | self._rst.storage,
+                    i_RST    = ResetSignal("sys") | self._rst.storage,
                     i_CLK    = ClockSignal(ddr_clk),
-                    i_CLKDIV = ClockSignal(),
+                    i_CLKDIV = ClockSignal("sys"),
                     **{f"i_D{n+1}": (0b10101010 >> n) & 0b1 for n in range(8)},
                     o_OQ     = sd_clk_se_nodelay,
                     i_OCE    = 1,
@@ -143,7 +146,7 @@ class S7DDRPHY(Module, AutoCSR):
                         p_REFCLK_FREQUENCY      = iodelay_clk_freq/1e6,
                         p_ODELAY_TYPE           = "VARIABLE",
                         p_ODELAY_VALUE          = 0,
-                        i_C        = ClockSignal(),
+                        i_C        = ClockSignal("sys"),
                         i_LD       = self._cdly_rst.re | self._rst.storage,
                         i_LDPIPEEN = 0,
                         i_CE       = self._cdly_inc.re,
@@ -159,18 +162,23 @@ class S7DDRPHY(Module, AutoCSR):
 
             # Commands -----------------------------------------------------------------------------
             commands = {
-                "a"    : "address",
-                "ba"   : "bank"   ,
-                "ras_n": "ras_n"  ,
-                "cas_n": "cas_n"  ,
-                "we_n" : "we_n"   ,
-                "cke"  : "cke"    ,
-                "odt"  : "odt"    ,
+                # Pad name: (DFI name,   Pad type (required or optional))
+                "reset_n" : ("reset_n", "optional"),
+                "cs_n"    : ("cs_n",    "optional"),
+                "a"       : ("address", "required"),
+                "ba"      : ("bank"   , "required"),
+                "ras_n"   : ("ras_n"  , "required"),
+                "cas_n"   : ("cas_n"  , "required"),
+                "we_n"    : ("we_n"   , "required"),
+                "cke"     : ("cke"    , "optional"),
+                "odt"     : ("odt"    , "optional"),
             }
-            if hasattr(pads, "reset_n"): commands.update({"reset_n" : "reset_n"})
-            if hasattr(pads, "cs_n")   : commands.update({"cs_n"    : "cs_n"})
-            for pad_name, dfi_name in commands.items():
-                pad = getattr(pads, pad_name)
+            for pad_name, (dfi_name, pad_type) in commands.items():
+                pad = getattr(pads, pad_name, None)
+                if (pad is None):
+                    if (pad_type == "required"):
+                        raise ValueError(f"DRAM pad {pad_name} required but not found in pads.")
+                    continue
                 for i in range(len(pad)):
                     oq  = Signal()
                     self.specials += Instance("OSERDESE2",
@@ -179,9 +187,9 @@ class S7DDRPHY(Module, AutoCSR):
                         p_TRISTATE_WIDTH = 1,
                         p_DATA_RATE_OQ   = "DDR",
                         p_DATA_RATE_TQ   = "BUF",
-                        i_RST    = ResetSignal() | self._rst.storage,
+                        i_RST    = ResetSignal("sys") | self._rst.storage,
                         i_CLK    = ClockSignal(ddr_clk),
-                        i_CLKDIV = ClockSignal(),
+                        i_CLKDIV = ClockSignal("sys"),
                         **{f"i_D{n+1}": getattr(dfi.phases[n//2], dfi_name)[i] for n in range(8)},
                         i_OCE    = 1,
                         o_OQ     = oq if with_odelay else pad[i],
@@ -196,7 +204,7 @@ class S7DDRPHY(Module, AutoCSR):
                             p_REFCLK_FREQUENCY      = iodelay_clk_freq/1e6,
                             p_ODELAY_TYPE           = "VARIABLE",
                             p_ODELAY_VALUE          = 0,
-                            i_C        = ClockSignal(),
+                            i_C        = ClockSignal("sys"),
                             i_LD       = self._cdly_rst.re | self._rst.storage,
                             i_LDPIPEEN = 0,
                             i_CE       = self._cdly_inc.re,
@@ -234,9 +242,9 @@ class S7DDRPHY(Module, AutoCSR):
                 p_TRISTATE_WIDTH = 1,
                 p_DATA_RATE_OQ   = "DDR",
                 p_DATA_RATE_TQ   = "BUF",
-                i_RST    = ResetSignal() | self._rst.storage,
+                i_RST    = ResetSignal("sys") | self._rst.storage,
                 i_CLK    = ClockSignal(ddr_clk) if with_odelay else ClockSignal(ddr_clk+"_dqs"),
-                i_CLKDIV = ClockSignal(),
+                i_CLKDIV = ClockSignal("sys"),
                 **{f"i_D{n+1}": dqs_bitslip.o[n] for n in range(8)},
                 i_OCE    = 1,
                 o_OFB    = dqs_o_no_delay if with_odelay else Signal(),
@@ -255,7 +263,7 @@ class S7DDRPHY(Module, AutoCSR):
                     p_PIPE_SEL              = "FALSE",
                     p_ODELAY_TYPE           = "VARIABLE",
                     p_ODELAY_VALUE          = half_sys8x_taps,
-                    i_C        = ClockSignal(),
+                    i_C        = ClockSignal("sys"),
                     i_LD       = (self._dly_sel.storage[i] & self._wdly_dqs_rst.re) | self._rst.storage,
                     i_CE       = self._dly_sel.storage[i] & self._wdly_dqs_inc.re,
                     i_LDPIPEEN = 0,
@@ -286,9 +294,9 @@ class S7DDRPHY(Module, AutoCSR):
                     p_TRISTATE_WIDTH = 1,
                     p_DATA_RATE_OQ   = "DDR",
                     p_DATA_RATE_TQ   = "BUF",
-                    i_RST    = ResetSignal() | self._rst.storage,
+                    i_RST    = ResetSignal("sys") | self._rst.storage,
                     i_CLK    = ClockSignal(ddr_clk),
-                    i_CLKDIV = ClockSignal(),
+                    i_CLKDIV = ClockSignal("sys"),
                     **{f"i_D{n+1}": dm_o_bitslip.o[n] for n in range(8)},
                     i_OCE    = 1,
                     o_OQ     = dm_o_nodelay if with_odelay else pads.dm[i],
@@ -303,7 +311,7 @@ class S7DDRPHY(Module, AutoCSR):
                         p_REFCLK_FREQUENCY      = iodelay_clk_freq/1e6,
                         p_ODELAY_TYPE           = "VARIABLE",
                         p_ODELAY_VALUE          = 0,
-                        i_C        = ClockSignal(),
+                        i_C        = ClockSignal("sys"),
                         i_LD       = (self._dly_sel.storage[i] & self._wdly_dq_rst.re) | self._rst.storage,
                         i_LDPIPEEN = 0,
                         i_CE       = self._dly_sel.storage[i] & self._wdly_dq_inc.re,
@@ -336,9 +344,9 @@ class S7DDRPHY(Module, AutoCSR):
                 p_TRISTATE_WIDTH = 1,
                 p_DATA_RATE_OQ   = "DDR",
                 p_DATA_RATE_TQ   = "BUF",
-                i_RST    = ResetSignal() | self._rst.storage,
+                i_RST    = ResetSignal("sys") | self._rst.storage,
                 i_CLK    = ClockSignal(ddr_clk),
-                i_CLKDIV = ClockSignal(),
+                i_CLKDIV = ClockSignal("sys"),
                 **{f"i_D{n+1}": dq_o_bitslip.o[n] for n in range(8)},
                 i_TCE    = 1,
                 i_T1     = ~dq_oe_delay.output,
@@ -358,10 +366,10 @@ class S7DDRPHY(Module, AutoCSR):
                 p_DATA_RATE      = "DDR",
                 p_NUM_CE         = 1,
                 p_IOBDELAY       = "IFD",
-                i_RST     = ResetSignal() | self._rst.storage,
+                i_RST     = ResetSignal("sys") | self._rst.storage,
                 i_CLK     = ClockSignal(ddr_clk),
                 i_CLKB    = ~ClockSignal(ddr_clk),
-                i_CLKDIV  = ClockSignal(),
+                i_CLKDIV  = ClockSignal("sys"),
                 i_BITSLIP = 0,
                 i_CE1     = 1,
                 i_DDLY    = dq_i_delayed,
@@ -379,7 +387,7 @@ class S7DDRPHY(Module, AutoCSR):
                     p_PIPE_SEL              = "FALSE",
                     p_ODELAY_TYPE           = "VARIABLE",
                     p_ODELAY_VALUE          = 0,
-                    i_C        = ClockSignal(),
+                    i_C        = ClockSignal("sys"),
                     i_LD       = (self._dly_sel.storage[i//8] & self._wdly_dq_rst.re)| self._rst.storage,
                     i_LDPIPEEN = 0,
                     i_CE       = self._dly_sel.storage[i//8] & self._wdly_dq_inc.re,
@@ -396,7 +404,7 @@ class S7DDRPHY(Module, AutoCSR):
                 p_PIPE_SEL              = "FALSE",
                 p_IDELAY_TYPE           = "VARIABLE",
                 p_IDELAY_VALUE          = 0,
-                i_C        = ClockSignal(),
+                i_C        = ClockSignal("sys"),
                 i_LD       = (self._dly_sel.storage[i//8] & self._rdly_dq_rst.re) | self._rst.storage,
                 i_LDPIPEEN = 0,
                 i_CE       = self._dly_sel.storage[i//8] & self._rdly_dq_inc.re,
