@@ -28,6 +28,8 @@ class USDDRPHY(Module, AutoCSR):
         memtype          = "DDR3",
         sys_clk_freq     = 100e6,
         iodelay_clk_freq = 200e6,
+        cl               = None,
+        cwl              = None,
         cmd_latency      = 0,
         cmd_delay        = None,
         is_rdimm         = False):
@@ -49,7 +51,8 @@ class USDDRPHY(Module, AutoCSR):
         if phytype == "USDDRPHY":  assert iodelay_clk_freq >= 200e6
         if phytype == "USPDDRPHY": assert iodelay_clk_freq >= 300e6
 
-        cl, cwl         = get_cl_cw(memtype, tck)
+        cl              = get_default_cl( memtype, tck) if cl  is None else cl
+        cwl             = get_default_cwl(memtype, tck) if cwl is None else cwl
         cl_sys_latency  = get_sys_latency(nphases, cl)
         cwl_sys_latency = get_sys_latency(nphases, cwl)
         rdphase         = get_sys_phase(nphases, cl_sys_latency,   cl + cmd_latency)
@@ -80,6 +83,7 @@ class USDDRPHY(Module, AutoCSR):
         self._wdly_dq_inc         = CSR()
         self._wdly_dqs_rst        = CSR()
         self._wdly_dqs_inc        = CSR()
+        self._wdly_dqs_inc_count  = CSRStatus(9)
 
         self._wdly_dq_bitslip_rst = CSR()
         self._wdly_dq_bitslip     = CSR()
@@ -138,9 +142,9 @@ class USDDRPHY(Module, AutoCSR):
                     p_IS_RST_INVERTED    = 0,
                     p_IS_CLK_INVERTED    = 0,
                     p_IS_CLKDIV_INVERTED = 0,
-                    i_RST    = ResetSignal() | self._rst.storage,
+                    i_RST    = ResetSignal("ic") | self._rst.storage,
                     i_CLK    = ClockSignal("sys4x"),
-                    i_CLKDIV = ClockSignal(),
+                    i_CLKDIV = ClockSignal("sys"),
                     i_D      = 0b10101010,
                     o_OQ     = clk_o_nodelay,
                 ),
@@ -152,8 +156,8 @@ class USDDRPHY(Module, AutoCSR):
                     p_DELAY_FORMAT     = "TIME",
                     p_DELAY_TYPE       = "VARIABLE",
                     p_DELAY_VALUE      = 0,
-                    i_RST          = self._cdly_rst.re | self._rst.storage,
-                    i_CLK          = ClockSignal(),
+                    i_RST          = ResetSignal("ic") | self._cdly_rst.re | self._rst.storage,
+                    i_CLK          = ClockSignal("sys"),
                     i_EN_VTC       = self._en_vtc.storage,
                     i_CE           = self._cdly_inc.re,
                     i_INC          = 1,
@@ -171,19 +175,24 @@ class USDDRPHY(Module, AutoCSR):
             # Commands -----------------------------------------------------------------------------
             pads_ba = Signal(bankbits)
             commands = {
-                "a"     : "address",
-                pads_ba : "bank",
-                "ras_n" : "ras_n"  ,
-                "cas_n" : "cas_n"  ,
-                "we_n"  : "we_n"   ,
-                "cke"   : "cke"    ,
-                "odt"   : "odt"    ,
+                # Pad name: (DFI name,   Pad type (required or optional))
+                "reset_n" : ("reset_n", "optional"),
+                "cs_n"    : ("cs_n",    "optional"),
+                "a"       : ("address", "required"),
+                pads_ba   : ("bank"   , "required"),
+                "ras_n"   : ("ras_n"  , "required"),
+                "cas_n"   : ("cas_n"  , "required"),
+                "we_n"    : ("we_n"   , "required"),
+                "cke"     : ("cke"    , "optional"),
+                "odt"     : ("odt"    , "optional"),
+                "act_n"   : ("act_n",   "optional"),
             }
-            if hasattr(pads, "reset_n"): commands.update({"reset_n" : "reset_n"})
-            if hasattr(pads, "cs_n")   : commands.update({"cs_n"    : "cs_n"})
-            if hasattr(pads, "act_n")  : commands.update({"act_n"   : "act_n"})
-            for pad_name, dfi_name in commands.items():
-                pad = pad_name if isinstance(pad_name, Signal) else getattr(pads, pad_name)
+            for pad_name, (dfi_name, pad_type) in commands.items():
+                pad = pad_name if isinstance(pad_name, Signal) else getattr(pads, pad_name, None)
+                if (pad is None):
+                    if (pad_type == "required"):
+                        raise ValueError(f"DRAM pad {pad_name} required but not found in pads.")
+                    continue
                 for i in range(len(pad)):
                     o_nodelay = Signal()
                     self.specials += [
@@ -194,9 +203,9 @@ class USDDRPHY(Module, AutoCSR):
                             p_IS_RST_INVERTED    = 0,
                             p_IS_CLK_INVERTED    = 0,
                             p_IS_CLKDIV_INVERTED = 0,
-                            i_RST    = ResetSignal() | self._rst.storage,
+                            i_RST    = ResetSignal("ic") | self._rst.storage,
                             i_CLK    = ClockSignal("sys4x"),
-                            i_CLKDIV = ClockSignal(),
+                            i_CLKDIV = ClockSignal("sys"),
                             i_D      = Cat(*[getattr(dfi.phases[n//2], dfi_name)[i] for n in range(8)]),
                             o_OQ     = o_nodelay,
                         ),
@@ -208,8 +217,8 @@ class USDDRPHY(Module, AutoCSR):
                             p_DELAY_FORMAT     = "TIME",
                             p_DELAY_TYPE       = "VARIABLE",
                             p_DELAY_VALUE      = 0,
-                            i_RST     = self._cdly_rst.re | self._rst.storage,
-                            i_CLK     = ClockSignal(),
+                            i_RST     = ResetSignal("ic") | self._cdly_rst.re | self._rst.storage,
+                            i_CLK     = ClockSignal("sys"),
                             i_EN_VTC  = self._en_vtc.storage,
                             i_CE      = self._cdly_inc.re,
                             i_INC     = 1,
@@ -238,18 +247,6 @@ class USDDRPHY(Module, AutoCSR):
         self.submodules += dqs_oe_delay, dqs_pattern
         self.comb += dqs_oe_delay.input.eq(dqs_preamble | dqs_oe | dqs_postamble)
         for i in range(databits//8):
-            if i == 0:
-                # Store initial DQS DELAY_VALUE (in taps) to be able to reload DELAY_VALUE after reset.
-                dqs_taps       = Signal(9)
-                dqs_taps_timer = WaitTimer(2**16)
-                dqs_taps_done  = Signal()
-                self.submodules += dqs_taps_timer
-                self.comb += dqs_taps_timer.wait.eq(~dqs_taps_done)
-                self.sync += \
-                    If(dqs_taps_timer.done,
-                        dqs_taps_done.eq(1),
-                        self._half_sys8x_taps.status.eq(dqs_taps)
-                    )
             dqs_bitslip    = BitSlip(8,
                 i      = dqs_pattern.o,
                 rst    = (self._dly_sel.storage[i] & self._wdly_dq_bitslip_rst.re) | self._rst.storage,
@@ -272,14 +269,13 @@ class USDDRPHY(Module, AutoCSR):
                         p_IS_RST_INVERTED    = 0,
                         p_IS_CLK_INVERTED    = 0,
                         p_IS_CLKDIV_INVERTED = 0,
-                        i_RST    = ResetSignal() | self._rst.storage,
+                        i_RST    = ResetSignal("ic") | self._rst.storage,
                         i_CLK    = ClockSignal("sys4x"),
-                        i_CLKDIV = ClockSignal(),
+                        i_CLKDIV = ClockSignal("sys"),
                         i_T      = ~dqs_oe_delay.output,
                         i_D      = dqs_bitslip.o,
                         o_OQ     = dqs_nodelay,
                         o_T_OUT  = dqs_t,
-
                     ),
                     Instance("ODELAYE3",
                         p_SIM_DEVICE       = device,
@@ -291,14 +287,14 @@ class USDDRPHY(Module, AutoCSR):
                         p_DELAY_FORMAT     = "TIME",
                         p_DELAY_TYPE       = "VARIABLE",
                         p_DELAY_VALUE      = int(tck*1e12/4),
-                        i_RST         = (self._dly_sel.storage[i] & self._wdly_dqs_rst.re) | self._rst.storage,
-                        i_CLK         = ClockSignal(),
-                        i_EN_VTC      = self._en_vtc.storage,
-                        i_CE          = self._dly_sel.storage[i] & self._wdly_dqs_inc.re,
-                        i_INC         = 1,
-                        o_CNTVALUEOUT = Signal(9) if i != 0 or j != 0 else dqs_taps,
-                        i_ODATAIN     = dqs_nodelay,
-                        o_DATAOUT     = dqs_delayed,
+                        o_CNTVALUEOUT      = self._half_sys8x_taps.status if (i == 0) and (j == 0) else Signal(),
+                        i_RST              = ResetSignal("ic"),
+                        i_CLK              = ClockSignal("sys"),
+                        i_EN_VTC           = self._en_vtc.storage,
+                        i_CE               = self._dly_sel.storage[i] & self._wdly_dqs_inc.re,
+                        i_INC              = 1,
+                        i_ODATAIN          = dqs_nodelay,
+                        o_DATAOUT          = dqs_delayed,
                     ),
                     Instance("IOBUFDSE3",
                         i_I    = dqs_delayed,
@@ -307,6 +303,9 @@ class USDDRPHY(Module, AutoCSR):
                         io_IOB = dqs_n,
                     )
                 ]
+                wdly_dqs_inc_count = Signal(9)
+                self.sync += If(self._dly_sel.storage[i] & self._wdly_dqs_inc.re, wdly_dqs_inc_count.eq(wdly_dqs_inc_count + 1))
+                self.comb += If(self._dly_sel.storage[i], self._wdly_dqs_inc_count.status.eq(wdly_dqs_inc_count))
 
         # DM ---------------------------------------------------------------------------------------
         for i in range(databits//8):
@@ -329,9 +328,9 @@ class USDDRPHY(Module, AutoCSR):
                         p_IS_RST_INVERTED    = 0,
                         p_IS_CLK_INVERTED    = 0,
                         p_IS_CLKDIV_INVERTED = 0,
-                        i_RST    = ResetSignal() | self._rst.storage,
+                        i_RST    = ResetSignal("ic") | self._rst.storage,
                         i_CLK    = ClockSignal("sys4x"),
-                        i_CLKDIV = ClockSignal(),
+                        i_CLKDIV = ClockSignal("sys"),
                         i_D      = dm_o_bitslip.o,
                         o_OQ     = dm_o_nodelay,
                     ),
@@ -345,9 +344,9 @@ class USDDRPHY(Module, AutoCSR):
                         p_DELAY_FORMAT     = "TIME",
                         p_DELAY_TYPE       = "VARIABLE",
                         p_DELAY_VALUE      = 0,
-                        i_RST     = (self._dly_sel.storage[i] & self._wdly_dq_rst.re) | self._rst.storage,
+                        i_RST     = ResetSignal("ic") | (self._dly_sel.storage[i] & self._wdly_dq_rst.re) | self._rst.storage,
                         i_EN_VTC  = self._en_vtc.storage,
-                        i_CLK     = ClockSignal(),
+                        i_CLK     = ClockSignal("sys"),
                         i_CE      = self._dly_sel.storage[i] & self._wdly_dq_inc.re,
                         i_INC     = 1,
                         i_ODATAIN = dm_o_nodelay,
@@ -379,9 +378,9 @@ class USDDRPHY(Module, AutoCSR):
                 p_IS_RST_INVERTED    = 0,
                 p_IS_CLK_INVERTED    = 0,
                 p_IS_CLKDIV_INVERTED = 0,
-                i_RST    = ResetSignal() | self._rst.storage,
+                i_RST    = ResetSignal("ic") | self._rst.storage,
                 i_CLK    = ClockSignal("sys4x"),
-                i_CLKDIV = ClockSignal(),
+                i_CLKDIV = ClockSignal("sys"),
                 i_D      = dq_o_bitslip.o,
                 i_T      = ~dq_oe_delay.output,
                 o_OQ     = dq_o_nodelay,
@@ -397,10 +396,10 @@ class USDDRPHY(Module, AutoCSR):
                 p_IS_CLK_INVERTED   = 0,
                 p_IS_CLK_B_INVERTED = 1,
                 p_DATA_WIDTH        = 8,
-                i_RST        = ResetSignal(),
+                i_RST        = ResetSignal("ic") | self._rst.storage,
                 i_CLK        = ClockSignal("sys4x"),
                 i_CLK_B      = ClockSignal("sys4x"), # locally inverted
-                i_CLKDIV     = ClockSignal(),
+                i_CLKDIV     = ClockSignal("sys"),
                 i_D          = dq_i_delayed,
                 i_FIFO_RD_EN = 0,
                 o_Q          = dq_i_bitslip.i,
@@ -417,8 +416,8 @@ class USDDRPHY(Module, AutoCSR):
                 p_DELAY_FORMAT     = "TIME",
                 p_DELAY_TYPE       = "VARIABLE",
                 p_DELAY_VALUE      = 0,
-                i_RST     = (self._dly_sel.storage[i//8] & self._wdly_dq_rst.re) | self._rst.storage,
-                i_CLK     = ClockSignal(),
+                i_RST     = ResetSignal("ic") | (self._dly_sel.storage[i//8] & self._wdly_dq_rst.re) | self._rst.storage,
+                i_CLK     = ClockSignal("sys"),
                 i_EN_VTC  = self._en_vtc.storage,
                 i_CE      = self._dly_sel.storage[i//8] & self._wdly_dq_inc.re,
                 i_INC     = 1,
@@ -436,8 +435,8 @@ class USDDRPHY(Module, AutoCSR):
                 p_DELAY_SRC        = "IDATAIN",
                 p_DELAY_TYPE       = "VARIABLE",
                 p_DELAY_VALUE      = 0,
-                i_RST     = (self._dly_sel.storage[i//8] & self._rdly_dq_rst.re) | self._rst.storage,
-                i_CLK     = ClockSignal(),
+                i_RST     = ResetSignal("ic") | (self._dly_sel.storage[i//8] & self._rdly_dq_rst.re) | self._rst.storage,
+                i_CLK     = ClockSignal("sys"),
                 i_EN_VTC  = self._en_vtc.storage,
                 i_CE      = self._dly_sel.storage[i//8] & self._rdly_dq_inc.re,
                 i_INC     = 1,
